/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.util.event.impl;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.Semaphore;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.TimeUnit;

import org.ow2.util.event.api.EventPriority;
import org.ow2.util.event.api.IEvent;
import org.ow2.util.event.api.IEventDispatcher;
import org.ow2.util.event.api.IEventListener;
import org.ow2.util.event.api.IEventToken;

/**
 * Define methods needed to register listeners and dispatch events.
 * @author missonng
 */
public class EventDispatcher implements IEventDispatcher {
    /**
     * A specialized queue synchronized give job to workers.
     * @author missonng
     */
    private final class EventQueue {
        /**
         * A SynchronousQueue for worker to wait jobs.
         */
        private final SynchronousQueue<EventToken> synchronousQueue = new SynchronousQueue<EventToken>(true);

        /**
         * A ConcurrentLinkedQueue to store jobs when no worker is available.
         */
        private final ConcurrentLinkedQueue<EventToken> concurrentLinkedQueue = new ConcurrentLinkedQueue<EventToken>();

        /**
         * Try to give a job to a worker directly through the SynchronousQueue.<br>
         * Store it in the ConcurrentLinkedQueue if no worker is available.
         * @param token The token to process.
         */
        public void put(final EventToken token) {
            if (!synchronousQueue.offer(token)) {
                concurrentLinkedQueue.offer(token);
            }
        }

        /**
         * Try to take a job from the ConcurrentLinkedQueue.<br>
         * Wait on the synchronousQueue if no job is available.
         * @return The runnable to execute.
         */
        public EventToken take() {
            while (true) {
                try {
                    EventToken token = concurrentLinkedQueue.poll();
                    if (token != null) {
                        return token;
                    }
                    return synchronousQueue.take();
                } catch (InterruptedException e) {
                    continue;
                }
            }
        }
    }

    /**
     * A thread speciliazed to process event tokens.<br>
     * Workers get tokens to process from the dispatcher queue.<br>
     * A worker dies when the special runnable KILL is taken from the queue.<br>
     * This method avoid complex synchronization to kill workers properly.
     * @author missonng
     */
    private final class EventWorker extends Thread {
        /**
         * The worker main method.
         */
        @Override
        public void run() {
            final EventQueue queue = EventDispatcher.this.queue;
            final IEventToken kill = EventDispatcher.this.kill;
            while (true) {
                final EventToken token = queue.take();

                if (token == kill) {
                    return;
                }

                token.lock.acquireUninterruptibly();
                if (token.killed) {
                    token.lock.release();
                    continue;
                }
                token.workers.add(this);
                token.lock.release();

                while (true) {
                    final IEventListener listener = token.listeners.poll();

                    if (listener == null) {
                        break;
                    }

                    final EventPriority[] priorities = EventPriority.values();
                    final EventPriority currPrio = token.priorities.get(listener);
                    final EventPriority prevPrio = currPrio.ordinal() > 0 ? priorities[currPrio.ordinal() - 1] : null;
                    final Semaphore currSemaphore = token.semaphores.get(currPrio);
                    final Semaphore prevSemaphore = token.semaphores.get(prevPrio);
                    final int prevCount = token.counts.get(prevPrio);
                    prevSemaphore.acquireUninterruptibly(prevCount);
                    prevSemaphore.release(prevCount);
                    try {
                        listener.handle(token.event);
                    } catch (Exception ex) {
                        continue;
                    } finally {
                        currSemaphore.release();
                    }
                }

                token.lock.acquireUninterruptibly();
                token.workers.remove(this);
                token.lock.release();
            }
        }
    }

    /**
     * Define methods need to manage events processing.
     * @author missonng
     */
    private final class EventToken implements IEventToken {
        /**
         * The event to manage.
         */
        private IEvent event;

        /**
         * The token state.
         */
        private boolean killed = false;

        /**
         * A semaphore to protect the token state and workers list.
         */
        private Semaphore lock = new Semaphore(1, true);

        /**
         * The token workers list.
         */
        private LinkedList<EventWorker> workers = new LinkedList<EventWorker>();

        /**
         * Map each priority to its number of listeners.
         */
        private HashMap<EventPriority, Integer> counts = new HashMap<EventPriority, Integer>();

        /**
         * Map each priority to its semaphore.
         */
        private HashMap<EventPriority, Semaphore> semaphores = new HashMap<EventPriority, Semaphore>();

        /**
         * Map each listener to its priority.
         */
        private HashMap<IEventListener, EventPriority> priorities = new HashMap<IEventListener, EventPriority>();

        /**
         * Listeners to process.
         */
        private ConcurrentLinkedQueue<IEventListener> listeners = new ConcurrentLinkedQueue<IEventListener>();

        /**
         * The event token constructor.<br>
         * Used to instanciate the kill token.
         */
        private EventToken() {

        }

        /**
         * The event token constructor.<br>
         * @param event The event to manage.
         */
        private EventToken(final IEvent event) {
            this.event = event;
            this.counts.put(null, 0);
            this.semaphores.put(null, new Semaphore(0, true));

            for (EventPriority priority : EventPriority.values()) {
                int count = 0;
                for (IEventListener listener : EventDispatcher.this.listeners.get(priority)) {
                    if (listener.accept(event)) {
                        this.priorities.put(listener, priority);
                        this.listeners.add(listener);
                        count++;
                    }
                }

                this.counts.put(priority, count);
                this.semaphores.put(priority, new Semaphore(0, true));
            }
        }

        /**
         * Terminate the event processing.
         */
        @SuppressWarnings("deprecation")
        public void terminate() {
            this.lock.acquireUninterruptibly();
            this.killed = true;
            for (EventWorker worker : this.workers) {
                worker.stop();
            }
            EventDispatcher.this.createWorkers(this.workers.size());
            this.lock.release();
        }

        /**
         * Allocate a new thread to process this event.
         */
        public void allocate() {
            EventDispatcher.this.queue.put(this);
        }

        /**
         * Wait for the event to be process.
         * @param timeout How long to wait for the event to be process.
         * @return True if the event has terminate its process before the timeout, false otherwise.
         * @throws InterruptedException If the thread has been interrupted.
         */
        public boolean join(final long timeout) throws InterruptedException {
            return join(EventPriority.values()[EventPriority.values().length - 1], timeout, false);
        }

        /**
         * Wait for the event to be process.
         * @param timeout How long to wait for the event to be process.
         * @param kill If true, kill the event processing if it not terminated before timeout.
         * @return True if the event has terminate its process before the timeout, false otherwise.
         * @throws InterruptedException If the thread has been interrupted.
         */
        public boolean join(final long timeout, final boolean kill) throws InterruptedException {
            return join(EventPriority.values()[EventPriority.values().length - 1], timeout, kill);
        }

        /**
         * Wait for the event to be process.
         * @param priority Only wait for the specified priority to be process.
         * @param timeout How long to wait for the event to be process.
         * @return True if the event has terminate its process before the timeout, false otherwise.
         * @throws InterruptedException If the thread has been interrupted.
         */
        public boolean join(final EventPriority priority, final long timeout) throws InterruptedException {
            return join(priority, timeout, false);
        }

        /**
         * Wait for the event to be process.
         * @param priority Only wait for the specified priority to be process.
         * @param timeout How long to wait for the event to be process.
         * @param kill If true, kill the event processing if it not terminated before timeout.
         * @return True if the event has terminate its process before the timeout, false otherwise.
         * @throws InterruptedException If the thread has been interrupted.
         */
        public boolean join(final EventPriority priority, final long timeout, final boolean kill) throws InterruptedException {
            final int count = this.counts.get(priority);
            final Semaphore semaphore = this.semaphores.get(priority);
            final boolean ok = semaphore.tryAcquire(count, timeout, TimeUnit.MILLISECONDS);
            if (ok) {
                semaphore.release(count);
            } else if (kill) {
                terminate();
            }
            return ok;
        }
    }

    /**
     * The dispatcher default size.
     */
    private static final int DEFAULT_SIZE = 1;

    /**
     * The listener map.<br>
     * Map each priority to a list of listeners.
     */
    private final HashMap<EventPriority, CopyOnWriteArrayList<IEventListener>> listeners;

    /**
     * The token queue.
     */
    private final EventQueue queue;

    /**
     * The special token used to kill workers.
     */
    private final EventToken kill;

    /**
     * The dispatcher state.
     */
    private boolean available;

    /**
     * The number of workers.
     */
    private int size;

    /**
     * The dispatcher constructor.
     */
    public EventDispatcher() {
        this.kill = new EventToken();
        this.queue = new EventQueue();

        this.available = false;

        this.size = DEFAULT_SIZE;

        this.listeners = new HashMap<EventPriority, CopyOnWriteArrayList<IEventListener>>();
        for (EventPriority priority : EventPriority.values()) {
            this.listeners.put(priority, new CopyOnWriteArrayList<IEventListener>());
        }
    }

    /**
     * Start the dispatcher.
     */
    public synchronized void start() {
        if (!this.available) {
            this.available = true;
            this.createWorkers(this.size);
        }
    }

    /**
     * Stop the dispatcher.
     */
    public synchronized void stop() {
        if (this.available) {
            this.available = false;
            this.destroyWorkers(this.size);
        }
    }

    /**
     * Get the dispatcher state.
     * @return The dispatcher state.
     */
    public synchronized boolean isAvailable() {
        return this.available;
    }

    /**
     * Get the number of workers.
     * @return The number of workers.
     */
    public synchronized int getSize() {
        return this.size;
    }

    /**
     * Set the number of workers.
     * @param size The number of workers.
     */
    public synchronized void setSize(final int size) {
        if (this.available) {
            final int diff = this.size - size;
            if (diff > 0) {
                destroyWorkers(diff);
            } else {
                createWorkers(-diff);
            }
        }
        this.size = size;
    }

    /**
     * Add a listener.
     * @param listener The listener to add.
     */
    public void addListener(final IEventListener listener) {
        this.listeners.get(listener.getPriority()).add(listener);
    }

    /**
     * Remove a listener.
     * @param listener The listener to remove.
     */
    public void removeListener(final IEventListener listener) {
        this.listeners.get(listener.getPriority()).remove(listener);
    }

    /**
     * Dispatch an event.
     * @param event The event to dispatch.
     * @return A token to manage the event processing.
     */
    public IEventToken dispatch(final IEvent event) {
        return dispatch(event, Long.MAX_VALUE);
    }

    /**
     * Dispatch an event.
     * @param event The event to dispatch.
     * @param timeout The maximum time dispatch waits for synchronous listeners to process.
     * @return A token to manage the event processing.
     */
    public IEventToken dispatch(final IEvent event, final long timeout) {
        if (!this.available) {
            throw new IllegalStateException();
        }

        final long start = System.currentTimeMillis();

        EventToken token = new EventToken(event);
        token.allocate();

        while (System.currentTimeMillis() < start + timeout) {
            try {
                token.join(EventPriority.values()[EventPriority.NONE.ordinal() - 1],
                        start + timeout - System.currentTimeMillis());
            } catch (InterruptedException e) {
                continue;
            }
        }
        return token;
    }

    /**
     * A helper method to create workers.
     * @param nb The number of workers to create.
     */
    private void createWorkers(final int nb) {
        for (int i = 0; i < nb; i++) {
            new EventWorker().start();
        }
    }

    /**
     * A helper method to destroy workers.
     * @param nb The number of workers to destroy.
     */
    private void destroyWorkers(final int nb) {
        final EventToken kill = this.kill;
        final EventQueue queue = this.queue;
        for (int i = 0; i < nb; i++) {
            queue.put(kill);
        }
    }
}
