/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.util.event.impl;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.ow2.util.event.api.IEvent;
import org.ow2.util.event.api.IEventDispatcher;
import org.ow2.util.event.api.IEventListener;
import org.ow2.util.event.api.IEventProvider;
import org.ow2.util.event.api.IEventService;
import org.ow2.util.event.api.IEventToken;
import org.ow2.util.event.api.exception.EventFilterMalformed;
import org.ow2.util.event.api.exception.EventIdAlreadyExists;
import org.ow2.util.event.api.exception.EventIdMalformed;
import org.ow2.util.event.api.exception.EventIdNotExists;

/**
 * Define a root event dispatcher, and manage a provider tree using a hierarchical naming convention.
 * By default, created provider will use their parent event provider, but they can redefine their own one.
 * Listener are registered in the service using a filter on provider ids.<br>
 * The service use this filter to register the listener on all matching providers.
 * @author missonng
 */
public class EventService implements IEventService {
    /**
     * An inner implementation of IEventProvider.
     * @author missonng
     */
    private final class EventProvider implements IEventProvider {
        /**
         * The event provider id.
         */
        private String id;

        /**
         * The event provider dispatcher.
         */
        private IEventDispatcher dispatcher;

        /**
         * The event provider listeners.
         */
        private List<IEventListener> listeners;

        /**
         * The event provider parent.
         */
        private EventProvider parent;

        /**
         * The event provider childs.
         */
        private Map<String, EventProvider> childs;

        /**
         * The root event provider constructor.
         */
        private EventProvider() {
            synchronized (EventService.this) {
                this.id = "/";
                this.dispatcher = new EventDispatcher();
                this.listeners = new LinkedList<IEventListener>();
                this.parent = null;
                this.childs = new HashMap<String, EventProvider>();
            }
        }

        /**
         * The event provider constructor.
         * @param id The event provider id.
         * @throws EventIdMalformed If the id does not match the regexp : (/\w+)+
         * @throws EventIdAlreadyExists If a provider with this id already exists.
         */
        private EventProvider(final String id) throws EventIdMalformed, EventIdAlreadyExists {
            synchronized (EventService.this) {
                if (!id.matches("(/\\w+)+")) {
                    throw new EventIdMalformed();
                }

                String parentId = id.substring(0, id.lastIndexOf("/"));
                EventProvider parent = (parentId.equals("") ? rootProvider : (EventProvider) providers.get(parentId));

                this.id = id;
                this.parent = parent == null ? new EventProvider(parentId) : parent;
                this.childs = new HashMap<String, EventProvider>();
                this.dispatcher = this.parent.dispatcher;
                this.listeners = new LinkedList<IEventListener>();
                this.parent.childs.put(id, this);

                for (Entry<IEventListener, String> entry : EventService.this.filters.entrySet()) {
                    if (this.id.matches(entry.getValue())) {
                        addListener(entry.getKey());
                    }
                }
            }
        }

        /**
         * Destroy the event provider.
         */
        private void destroy() {
            synchronized (EventService.this) {
                for (IEventListener listener : this.listeners) {
                    this.dispatcher.removeListener(listener);
                }

                for (EventProvider child : this.childs.values()) {
                    child.destroy();
                }

                if (this.parent != null) {
                    this.parent.childs.remove(id);
                    this.parent = null;
                }

            }
        }

        /**
         * Get the event provider id.
         * @return The event provider id.
         */
        public String getId() {
            synchronized (EventService.this) {
                return id;
            }
        }

        /**
         * Add an event listener.
         * @param listener The event listener to add.
         */
        public void addListener(final IEventListener listener) {
            synchronized (EventService.this) {
                this.listeners.add(listener);
                this.dispatcher.addListener(listener);
            }
        }

        /**
         * Dispatch this event.
         * @param event The event to dispatch.
         * @return A token to manage the event processing.
         * @throws EventDispatcherNotAvailable If the dispatcher is not available.
         */
        public IEventToken dispatch(final IEvent event) {
            synchronized (EventService.this) {
                return this.dispatcher.dispatch(event);
            }
        }

        /**
         * Dispatch an event.
         * @param event The event to dispatch.
         * @param timeout The maximum time dispatch waits for synchronous listeners to process.
         * @return A token to manage the event processing.
         */
        public IEventToken dispatch(final IEvent event, final long timeout) {
            synchronized (EventService.this) {
                return this.dispatcher.dispatch(event, timeout);
            }
        }

        /**
         * Remove an event listener.
         * @param listener The event listener to remove.
         */
        public void removeListener(final IEventListener listener) {
            synchronized (EventService.this) {
                this.listeners.remove(listener);
                this.dispatcher.removeListener(listener);
            }
        }

        /**
         * Get the event dispatcher used by this event provider.
         * @return The event dispatcher use by this event provider.
         */
        public IEventDispatcher getDispatcher() {
            synchronized (EventService.this) {
                return this.dispatcher;
            }
        }

        /**
         * Set the event dispatcher used by this event provider.
         * @param dispatcher The event dispatcher use by this event provider.
         */
        public void setDispatcher(final IEventDispatcher dispatcher) {
            synchronized (EventService.this) {
                IEventDispatcher oldDispatcher = this.dispatcher;
                IEventDispatcher newDispatcher = dispatcher == null ? this.parent.dispatcher : dispatcher;
                this.dispatcher = newDispatcher;

                for (IEventListener listener : listeners) {
                    newDispatcher.addListener(listener);
                    oldDispatcher.removeListener(listener);
                }

                for (EventProvider child : this.childs.values()) {
                    if (child.dispatcher == oldDispatcher) {
                        child.setDispatcher(newDispatcher);
                    }
                }
            }
        }
    }

    private boolean available;

    /**
     * The root event provider.
     */
    private EventProvider rootProvider;

    /**
     * The root event dispatcher.
     */
    private EventDispatcher rootDispatcher;

    /**
     * The service managed event providers.
     */
    private HashMap<String, EventProvider> providers;

    /**
     * The service managed event listeners.
     */
    private HashMap<IEventListener, String> filters;

    public EventService() {
        this.start();
    }

    /**
     * Create and return a new event provider with the specified id.
     * @param id The event provider id.
     * @return The event provider.
     * @throws EventIdMalformed If the id does not match the regexp : (/\w+)+
     * @throws EventIdAlreadyExists If a provider with this id already exists.
     */
    public synchronized IEventProvider createProvider(final String id) throws EventIdMalformed, EventIdAlreadyExists {
        if (!this.available) {
            throw new IllegalStateException();
        }

        if (this.providers.containsKey(id)) {
            throw new EventIdAlreadyExists();
        } else {
            EventProvider provider = new EventProvider(id);
            this.providers.put(id, provider);
            return provider;
        }
    }

    /**
     * Destroy an new event provider with the specified id.
     * @param id The event provider id.
     * @throws EventIdNotExists If no provider with this id exists.
     */
    public synchronized void destroyProvider(final String id) throws EventIdNotExists {
        if (!this.available) {
            throw new IllegalStateException();
        }

        EventProvider provider = this.providers.remove(id);
        if (provider != null) {
            while (!this.providers.containsKey(provider.parent.id)) {
                provider = provider.parent;
            }

            provider.destroy();
        } else {
            throw new EventIdNotExists();
        }
    }

    /**
     * Register a listener on all provider matching the filter.
     * @param listener The listener to register.
     * @param filter The filter to apply.
     * @throws EventFilterMalformed If the filter does not match the regexp : (/[\w\*]+)+
     */
    public synchronized void registerListener(final IEventListener listener, final String filter) throws EventFilterMalformed {
        if (!this.available) {
            throw new IllegalStateException();
        }

        if (!filter.matches("(/[\\w\\*]+)+")) {
            throw new EventFilterMalformed();
        }

        String newregexp = filter.replaceAll("\\*\\*", "[/\\\\w]#").replaceAll("\\*", "\\\\w#").replaceAll("#", "*");
        String oldregexp = filters.put(listener, newregexp);

        for (Entry<String, EventProvider> entry : this.providers.entrySet()) {
            if (oldregexp != null && entry.getKey().matches(oldregexp)) {
                entry.getValue().removeListener(listener);
            }

            if (entry.getKey().matches(newregexp)) {
                entry.getValue().addListener(listener);
            }
        }
    }

    /**
     * Unregister an event listener.
     * @param listener The event listener to unregister.
     */
    public synchronized void unregisterListener(final IEventListener listener) {
        if (!this.available) {
            throw new IllegalStateException();
        }

        String oldregexp = filters.remove(listener);
        if (oldregexp != null) {
            for (Entry<String, EventProvider> entry : this.providers.entrySet()) {
                if (entry.getKey().matches(oldregexp)) {
                    entry.getValue().removeListener(listener);
                }
            }
        }
    }

    /**
     * Start the service.
     */
    public synchronized void start() {
        if (!this.available) {
            this.available = true;
            rootDispatcher = new EventDispatcher();
            rootDispatcher.start();
            rootProvider = new EventProvider();
            rootProvider.setDispatcher(rootDispatcher);
            providers = new HashMap<String, EventProvider>();
            providers.put("/", rootProvider);
            filters = new HashMap<IEventListener, String>();
        }
    }

    /**
     * Stop the service.
     */
    public synchronized void stop() {
        if (this.available) {
            this.available = false;
            filters.clear();
            filters = null;
            providers.clear();
            providers = null;
            rootProvider.destroy();
            rootProvider = null;
            rootDispatcher.stop();
            rootDispatcher = null;
        }
    }
}